/* hf_dialog.c - height field interface functions
 *
 * Copyright (C) 2001 Patrice St-Gelais
 *         patrstg@users.sourceforge.net
 *         www.oricom.ca/patrice.st-gelais
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "hf_dialog.h"
#include <sys/types.h>
#include <signal.h>

extern gchar *HF_OUTPUT_FOR_RENDERING, *OUTPUT_PREFIX;

#define NBSIZES 6
void size_callb (gpointer data,GtkWidget *wdg);

omenu_list_type sizes_list[NBSIZES] = {
	NULL,128,NULL,"128",size_callb,
	NULL,256,NULL,"256",size_callb,
	NULL,512,NULL,"512",size_callb,
	NULL,1024,NULL,"1024",size_callb,
	NULL,2048,NULL,"2048",size_callb,
	NULL,4096,NULL,"4096",size_callb,
};

// Toolbar icons for HF algorithm
#define NBHFTYPES 4
#include "../icons/uniform.xpm"
#include "../icons/subdiv1.xpm"
#include "../icons/subdiv2.xpm"
// #include "../icons/waves.xpm"
// #include "../icons/waves1.xpm"
#include "../icons/surfadd.xpm"

// Toolbar icons for main HF tools
#define NBHFTOOLS 2
#include "../icons/transform.xpm"
#include "../icons/pencil.xpm"
#include "../icons/run2.xpm"
#include "../icons/run2_with_choice.xpm"
#include "../icons/stop.xpm"

command_item_struct hftypes[NBHFTYPES] = {
"Type", "Uniform", "Uniform color",0, (gchar **) uniform_xpm, GDK_LEFT_PTR, hfuniform_callb, NULL, NULL,TRUE,
"Type", "Subdiv1", "Subdivision 1",0, (gchar **) subdiv1_xpm, GDK_LEFT_PTR, hfsubdiv1_callb,NULL, NULL,FALSE,
"Type", "Subdiv2", "Subdivision 2",0, (gchar **) subdiv2_xpm, GDK_LEFT_PTR, hfsubdiv2_callb, NULL, NULL,FALSE,
"Type", "SurfSum", "Progressive sum of surfaces",0, (gchar **) surfadd_xpm, GDK_LEFT_PTR, hfsurfadd_callb, NULL, NULL,FALSE
};

command_item_struct hftools[NBHFTOOLS] = {
"Edit", 	  "Transform", "Filters and other transformations", 0, (gchar **) transform_xpm, GDK_LEFT_PTR, transform_callb,NULL, NULL, TRUE,
"Edit", 	  "Draw", "Drawing tools", 0, (gchar **) pencil_xpm, GDK_PENCIL, draw_tools_callb,NULL, NULL,FALSE
// "Edit", 	  NULL, NULL, 0, NULL, 0, NULL,NULL, NULL, FALSE,	// Separator
// "Tools",  "Run", "Run - POVRAY", 'R', (gchar **) run_xpm, GDK_LEFT_PTR, run_callb,NULL, NULL, FALSE,
// "Tools",  "Stop", "Stop - POVRAY", 0, (gchar **) stop_xpm, GDK_LEFT_PTR, stop_callb,NULL, NULL, FALSE,
};

#define NB_SCENE_EXT 2
omenu_list_type scene_ext_list[NB_SCENE_EXT] = {
	NULL,0,NULL,gettext_noop("POV"),filetype_sel_callback_fn,
	NULL,0,NULL,gettext_noop("<no extension>"),filetype_sel_callback_fn,
};

void accept_fn (hf_wrapper_struct *hfw) {
	if (!hfw)
		return;
//	printf("********** ACCEPT_CALLB\n");
	hf_backup(hfw->hf_struct);
	commit_pending_record(hfw);
	if (hfw->hf_options->img->set_fn)
		(*hfw->hf_options->img->set_fn) ((gpointer)hfw->hf_options->img);
//	Check if recomputing min-max is necessary or can be done on demand
//	hf_min_max(hfw->hf_struct);
	gtk_widget_set_sensitive(GTK_WIDGET(hfw->hf_options->img->accept_wdg),FALSE);
	hfw->if_calculated = FALSE;
}

void reset_fn (hf_wrapper_struct *hfw) {
//	Cancels the current operation by copying the temp buffer (backuped)
// 	into the current (modified) one
//	printf("********** RESET_CALLB\n");
	if (!hfw)
		return;
	if (hfw->hf_options->img->set_fn)
		(*hfw->hf_options->img->set_fn) ((gpointer)hfw->hf_options->img);
	hf_restore (hfw->hf_struct);
	gtk_widget_set_sensitive(GTK_WIDGET(hfw->hf_options->img->accept_wdg),FALSE);
	draw_hf(hfw);
}

void record_action (hf_wrapper_struct *data, gchar *action_lbl) {
//	Record action in the standard undo / redo stack
	hf_wrapper_struct *hfw;
	hfw = (hf_wrapper_struct *) data;
// printf("RECORD ACTION: %s\n", _(action_lbl));
	if (hfw->type != HFMAIN)	// Don't record an HFPEN, for instance
		return;
	if (hfw->init_history_flag)
		(*hfw->init_history_flag) (hfw->window, hfw->to_register_in_history);
	hfw->last_action = _(action_lbl);
}

void set_commit_reset_fns (hf_options_struct *hfo,
	gpointer commit_fn,
	gpointer reset_fn,
	gpointer data) {

	// Set the functions and data to execute when we change context
	*hfo->commit_fn = commit_fn;
	*hfo->reset_fn = reset_fn;
	*((hf_wrapper_struct **) hfo->commit_data) = data;
}

void begin_pending_record (hf_wrapper_struct *hfw, gchar *action_lbl,
	gpointer commit_fn,
	gpointer reset_fn) {

//	Anticipation of an action recording in the standard undo / redo stack
//	The signal for actually recording the action in the stack would be
//	emitted by commit_pending_record, after, for instance, a context change
//	Added 2004-03: initializes the functions required to commit / reset
//				the changes when we change context

//	printf("BEGIN_PENDING_RECORD: %s\n",action_lbl);
	if (hfw->to_register_in_history)
		(*hfw->to_register_in_history) = TRUE;
	hfw->last_action = _(action_lbl);
	set_commit_reset_fns (hfw->hf_options, commit_fn, reset_fn, (gpointer) hfw);
}

void commit_pending_record (hf_wrapper_struct *data) {
//	Commits an anticipated action record, in the undo / redo stack
	hf_wrapper_struct *hfw;
	hfw = (hf_wrapper_struct *) data;
// 	printf("COMMIT_PENDING_RECORD\n");
	if (hfw->to_register_in_history)	//	Hoping the flag address exists!
	{
// 	printf("Commit_pending_record - to_register_in_history: %d\n",hfw->to_register_in_history);
		if (*hfw->to_register_in_history) //	Flag on?
		{
// 	printf("Flag_on YES\n");
			if (hfw->init_history_flag)	// Hoping the recording function exists!
			{
// 	printf("FN: %d; window: %d\n",hfw->init_history_flag, hfw->to_register_in_history);
				(*hfw->init_history_flag) (hfw->window, hfw->to_register_in_history);
			}
		}
	}
}

void hf_commit_or_reset (hf_options_struct *hfo) {
//	Commits or resets a change, with the functions and data given in hf_options
	if (hfo->commit_data && hfo->commit_fn && hfo->reset_fn)
		commit_or_reset (hfo->commit_fn,
			hfo->reset_fn,
			* (hf_wrapper_struct **) hfo->commit_data);
}

gint get_display_scale (gint hf_size) {
//	Guess display scale from HF size and current display size...
//	or take the value given in the RC file
	static gint def_size = 0;
	if (DOC_DISPLAY_SIZE) {
		if (hf_size>DOC_DISPLAY_SIZE)
			return log2i(hf_size / DOC_DISPLAY_SIZE);
		else 
			return 0;
	}
	else {
		if (!def_size)
			def_size =  hf_default_size();
		if (hf_size > def_size)
			return log2i(hf_size / def_size);
		else
			return 0;
	}
}

void size_callb (gpointer data,GtkWidget *wdg) {
	// Sets size when creating a new HF
	gint hf_size;
	static gchar area_label[12];
	omenu_struct_type *opt_menu;
	hf_wrapper_struct *hfw;
	opt_menu = (omenu_struct_type *) data;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) opt_menu->data;

// printf("HFWrapper adr in SIZE_CALLB: %d;  HFWrapper: %d\n",opt_menu->data,hfw);

	hf_size = opt_menu->list[opt_menu->current_index].num;
	if (!hfw)
		return;
	if (hf_size == hfw->hf_struct->max_x)
		//	No change!
		return;

	sprintf(area_label, "%d x %d", hf_size, hf_size);
	gtk_frame_set_label(GTK_FRAME(hfw->draw_area_frame), area_label);

	hf_free(hfw->hf_struct);
	hfw->hf_options->hf_size = hf_size;
	hfw->display_scale = get_display_scale(hf_size);
	hfw->hf_struct = (hf_struct_type *) hf_new(hf_size, hfw->display_scale);
	set_widget_display_scale(hfw);
	hfw->area_size = RIGHT_SHIFT(hf_size,hfw->display_scale);
	hfw->size_change_request = TRUE;
	
	gtk_widget_set_size_request (hfw->draw_area_frame, hfw->area_size+SCALE_PAD, hfw->area_size+SCALE_PAD);

// printf("HF_SIZE: %d; DISPLAY_SCALE: %d; AREA_SIZE: %d\n",hf_size, hfw->display_scale, hfw->area_size);
	gl_update_details_display (hfw->gl_preview, hf_size);
	gtk_drawing_area_size(GTK_DRAWING_AREA(hfw->area), 
		hfw->area_size, hfw->area_size);
}


GtkWidget *hf_types_toolbar_new(GtkWidget *window, GtkTooltips *tooltips, 
		GtkWidget** tb_widget, gpointer data) {
	GtkWidget *hbox;
	toolbar_struct *tb;
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(GTK_WIDGET(hbox));
	define_label_in_box("Type", hbox, FALSE, FALSE, DEF_PAD*0.5);
	// toolbar_struct is a bit overkill for this purpose...
	tb = toolbar_new(NBHFTYPES,
		hftypes,
		tooltips,
		window,
		data,
		GTK_ORIENTATION_HORIZONTAL,
		GTK_TOOLBAR_ICONS,
		TRUE);
	gtk_box_pack_start(GTK_BOX(hbox), tb->toolbarwdg, FALSE, FALSE, DEF_PAD*0.5);
	if (tb_widget)
		*tb_widget = tb->toolbarwdg;
	free(tb);
	return hbox;
}

GtkWidget *create_run_bar (GtkWidget *window, gpointer data) {
//	Code for creating the Povray run buttons
//	Used twice, for the creation dialog and for the tools dialog
	GtkWidget *hbox, *button, *dummy;
	GtkTooltips *tooltips;

	tooltips = gtk_tooltips_new();

	dummy = gtk_alignment_new(0.9,0.5,0,0);
	gtk_widget_show(GTK_WIDGET(dummy));

	hbox = gtk_hbox_new(FALSE, DEF_PAD);
	gtk_widget_show(hbox);

	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button),
		create_widget_from_xpm(window,run2_with_choice_xpm));
    	gtk_signal_connect (GTK_OBJECT (button), "clicked",
                               GTK_SIGNAL_FUNC( run_with_choice_callb),
                               data);
	gtk_tooltips_set_tip(GTK_TOOLTIPS(tooltips),button,
		_("Choose a scene (.pov) and render it"),_("Choose a scene (.pov) and render it"));
	gtk_container_add(GTK_CONTAINER(hbox),button);
	gtk_widget_show(button);

	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button),
		create_widget_from_xpm(window,run2_xpm));
    	gtk_signal_connect (GTK_OBJECT (button), "clicked",
                               GTK_SIGNAL_FUNC( run_callb),
                               data);
	gtk_tooltips_set_tip(GTK_TOOLTIPS(tooltips),button,
		_("Render the current scene (.pov)"),_("Render the current scene (.pov)"));
	gtk_container_add(GTK_CONTAINER(hbox),button);
	gtk_widget_show(button);

	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button),
		create_widget_from_xpm(window,stop_xpm));
    	gtk_signal_connect (GTK_OBJECT (button), "clicked",
                               GTK_SIGNAL_FUNC( stop_callb),
                               data);
	gtk_tooltips_set_tip(GTK_TOOLTIPS(tooltips),button,
		_("Stop the scene rendering"),_("Stop the scene rendering"));
	gtk_container_add(GTK_CONTAINER(hbox),button);
	gtk_widget_show(button);
	
	gtk_container_add(GTK_CONTAINER(dummy),hbox);

	return dummy;
}

GtkWidget * hf_creation_dialog_new(GtkWidget *window, GtkTooltips *tooltips, 
		gpointer data, gpointer widget_list_ptr) {
	// Initializes the dialog for creating height fields
	// Four parts, from top to bottom:
	//	1. Kind / type of HF
	//	2. Shape filters
	//	3. Generic options (size...)
	//	4. Options for the current HF
	// Returns a vbox container, allowing to add / substract particular controls
	// "window" only needed for heritance of characteristics in child toolbar(s)
	// "tooltips" needed for packing tooltips, could have only one for the whole app.
	// gpointer data:  is normally the address of the current hf_wrapper pointer 
	// (hf_wrapper_struct **)

	GtkWidget *dialog, *hbox, *vbox, *frame;
	gint default_hf_size;
	omenu_struct_type *opt_menu;
	hf_creation_wdg_struct *wdg_str;

	wdg_str = (hf_creation_wdg_struct *) malloc(sizeof(hf_creation_wdg_struct));

	*((hf_creation_wdg_struct **) widget_list_ptr) = wdg_str;

	frame = options_frame_new("HeightField");

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);

	//	1st generic option:  HF size - a power of 2
	//	Default:  some % of screen width
	
	hbox = gtk_hbox_new(FALSE, DEF_PAD);
	gtk_widget_show(hbox);
	define_label_in_box("Size", hbox, FALSE, FALSE, DEF_PAD);

	default_hf_size = hf_default_size();
// printf("HFW_ADR: %d;  HFW: %d;  DEFAULT_HF_SIZE: %d\n",data, *((hf_wrapper_struct**) data),default_hf_size);
	
	create_omenu_with_callb_data (&opt_menu, sizes_list,
		NBSIZES, 
		find_omenu_num(sizes_list, default_hf_size, NBSIZES),
		data);
	gtk_box_pack_start(GTK_BOX(hbox), opt_menu->omenu_ptr, FALSE, FALSE, 0);

	wdg_str->size_menu = opt_menu;

//	*********** Remember to check where / when to free opt_menu!

 	gtk_box_pack_start (GTK_BOX (hbox), create_run_bar(window,data), TRUE, TRUE, 0);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(vbox),
		hf_types_toolbar_new(window, tooltips, &wdg_str->hf_types_toolbar, data) ,
		FALSE, FALSE, 0);

	//	Other generic options ?

	gtk_container_add(GTK_CONTAINER(frame), vbox);
	dialog = gtk_vbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(dialog), frame, FALSE, FALSE, 0);
	gtk_widget_show(dialog);

	return dialog;
}

GtkWidget * hf_tools_dialog_new(GtkWidget *window, GtkTooltips *tooltips,
	gpointer hf_opt, gpointer data) {
	// Initializes the dialog for editing height fields
	// Two parts, from top to bottom:
	//	1. Kind of edition:  paint, transform, run / stop (not edition, but...)
	//	2. Toolbar / options for the current edit button
	// Returns a vbox container, allowing to add / substract particular controls
	// "window" only needed for heritance of characteristics in child toolbar(s)
	// ... could be the application main window
	// "tooltips" needed for packing tooltips, could have only one for the whole app.
	// gpointer data:  is normally the address of the current hf_wrapper pointer (hf_wrapper **)

	GtkWidget *dialog, *hbox, *tb, *tmp;
	hf_options_struct *hf_options;
	hf_options = (hf_options_struct *) hf_opt;

	dialog = gtk_vbox_new(FALSE, DEF_PAD);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(GTK_WIDGET(hbox));
	define_label_in_box("Edit", hbox, FALSE, FALSE, DEF_PAD*0.5);
	tb = standard_toolbar_new(NBHFTOOLS, hftools,
		tooltips,
		window,
		data,
		GTK_ORIENTATION_HORIZONTAL,
		GTK_TOOLBAR_ICONS,
		TRUE);
	gtk_box_pack_start(GTK_BOX(hbox), tb, FALSE, FALSE, DEF_PAD);
	gtk_box_pack_start(GTK_BOX(dialog), hbox, FALSE, FALSE, DEF_PAD);

	gtk_box_pack_start(GTK_BOX(hbox), create_run_bar(window,data),
		TRUE, TRUE, DEF_PAD);

	gtk_widget_show(dialog);

	//	Sub-dialogs creation

	hf_options->draw = draw_tools_new(NULL,hf_options->dist_matrix);
	hf_options->draw->draw_tools_dialog =
		draw_tools_dialog_new(window, tooltips, hf_options->draw, data);
	gtk_container_add(GTK_CONTAINER(dialog), hf_options->draw->draw_tools_dialog);

	tmp = img_process_dialog_new(window, tooltips, data);
	hf_options->img = img_struct_new(NULL, data);
	hf_options->img->img_dialog = tmp;
	gtk_container_add(GTK_CONTAINER(dialog), tmp);

	//	The default is the transform dialog

	hf_options->current_tools = hf_options->img->img_dialog;
	gtk_widget_show(hf_options->img->img_dialog);
	gtk_widget_hide(hf_options->draw->draw_tools_dialog);

	hf_options->context = TRANSFORM_CONTEXT;
	hf_options->current_cursor = hf_options->default_cursor;

//	Default img subdialog - brightness / contrast
	hf_options->img->br_contr_dialog = br_contr_dialog_new(hf_options->img, data);
	gtk_container_add(GTK_CONTAINER(hf_options->img->img_dialog),
		hf_options->img->br_contr_dialog );
	hf_options->img->current_subdialog = hf_options->img->br_contr_dialog;
	hf_options->img->accept_wdg = hf_options->img->br_contr_accept ;
	hf_options->img->set_fn = (gpointer) set_br_contr_defaults;

	hf_options->tools_dialog = dialog;

	return dialog;
}

void hf_commit_creation (gpointer data) {
//	Commit the filter
	filter_dialog_struct *fds;
	hf_wrapper_struct *hfw;
	hfw = (hf_wrapper_struct *) data;
// printf("HFW, HFW->creation_mode in hf_commit_creation: %d, %d\n",data,hfw->creation_mode);
// printf("HFW->window->window: %d->%d\n",hfw->window,hfw->window->window);
	hfw->creation_mode = FALSE;  // Supposed to be done in hf_wrapper_copy...
//	hfw->hf_options->current_calculation = NULL;
	hfw->apply_filter = FALSE;

//	Backup the filter definition and nullify it - we 1st commit the HF without the filter
	if (hfw->hf_options->fd_struct) {
		set_filter_defaults (hfw->hf_options->fd_struct);
	}
	
	record_action(hfw, hfw->hf_options->last_creation_action);
	if (hfw->hf_options->fd_struct) {
		fds = hfw->hf_options->fd_struct;
		if (fds->current_filter) {
		    filter_apply(hfw->hf_struct,
			fds->current_filter,
			hfw->hf_options->dist_matrix,
			fds->revert_filter,
			fds->filter_level,
			fds->merge_oper);
//	Now we commit the HF with the filter
		record_action(hfw,_("Shape filter"));
		}
	}
//	If we are in the shape filter subdialog in the tools dialog:
	if (hfw->hf_options->img)
		hfw->hf_options->fd_struct = hfw->hf_options->img->fd_struct;
	else
		hfw->hf_options->fd_struct = NULL;

//	"expose_event" seems required to display the cursor ?
	emit_expose_event(hfw->area);
}

// HF creation

void hfuniform_callb(GtkWidget *wdg, gpointer data) {
	// Display / connect callback for uniform hf (a simple scale from 0 to 0xFFFF)

	hf_wrapper_struct *hfw;

	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;

// printf("IN hfuniform_callb:  data: %d;  hfw->hf_options: %d\n", data, hfw->hf_options);

	if (!hfw)
		return;
	//  Options dialog
	if (!hfw->hf_options->uniform_dialog) {
		// NB hfw->hf_options->uniform_value allocated in hf_options_new()
		hfw->hf_options->uniform_dialog = uniform_dialog_new(data,
			*hfw->hf_options->uniform_value);
		gtk_container_add(GTK_CONTAINER(hfw->creation_container),
			hfw->hf_options->uniform_dialog);
	}
	if (hfw->hf_options->current_dialog)
		gtk_widget_hide(hfw->hf_options->current_dialog);
	gtk_widget_show(hfw->hf_options->uniform_dialog);
	hfw->hf_options->last_creation_action = _("Uniform");
	hfw->hf_options->current_dialog = hfw->hf_options->uniform_dialog;
	hfw->hf_options->current_calculation = (gpointer) uniform;
	hfw->hf_options->current_callb = hfuniform_callb;
	hfw->hf_options->current_options = (gpointer) hfw->hf_options->uniform_value;
	hfw->hf_options->flag_uniform_done = FALSE;
	gener_hf(hfw);
}

void hfsubdiv1_callb(GtkWidget *wdg, gpointer data) {
	// Display / connect callback
	// Simple subdivision, square / diamond dots

	hf_wrapper_struct *hfw;
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	if (!hfw)
		return;

	//  Options dialog - we create it at the 1st pass
	if (!hfw->hf_options->subdiv1_dialog) {
	//	We create the struct which will receive the data
		hfw->hf_options->subdiv1_options = subdiv1_opt_new();
	//	Dialog creation (with some defaults from subdiv1_opt_new())
		hfw->hf_options->subdiv1_dialog = subdiv1_dialog_new(data);
		gtk_container_add(GTK_CONTAINER(hfw->creation_container),
			hfw->hf_options->subdiv1_dialog);
	}
	if (hfw->hf_options->current_dialog)
		gtk_widget_hide(hfw->hf_options->current_dialog);
	gtk_widget_show(hfw->hf_options->subdiv1_dialog);
	hfw->hf_options->last_creation_action = _("Subdiv1");
	hfw->hf_options->current_dialog= hfw->hf_options->subdiv1_dialog;
	hfw->hf_options->current_calculation = (gpointer) subdiv1;
	hfw->hf_options->current_callb = hfsubdiv1_callb;
	hfw->hf_options->current_options = (gpointer) hfw->hf_options->subdiv1_options;
	gener_hf(hfw);
}

void hfsubdiv2_callb(GtkWidget *wdg, gpointer data) {
	// Display / connect callback
	// Simple subdivision, square / diamond dots - TEST

	hf_wrapper_struct *hfw;
	static gint frq = 50;
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	if (!hfw) 
		return;
	//  Options dialog - we create it at the 1st pass
	if (!hfw->hf_options->subdiv2_dialog) {
	//	We create the struct which will receive the data
		hfw->hf_options->subdiv2_options = subdiv2_opt_new(&frq,1);
	//	Dialog creation (with some defaults from subdiv2_opt_new())
		hfw->hf_options->subdiv2_dialog = subdiv2_dialog_new(data);
		gtk_container_add(GTK_CONTAINER(hfw->creation_container),
			hfw->hf_options->subdiv2_dialog);
	}
	if (hfw->hf_options->current_dialog)
		gtk_widget_hide(hfw->hf_options->current_dialog);
	gtk_widget_show(hfw->hf_options->subdiv2_dialog);
	hfw->hf_options->last_creation_action = _("Subdiv2");
	hfw->hf_options->current_dialog = hfw->hf_options->subdiv2_dialog;
	hfw->hf_options->current_calculation = (gpointer) subdiv2;
	hfw->hf_options->current_callb = hfsubdiv2_callb;
	hfw->hf_options->current_options = (gpointer) hfw->hf_options->subdiv2_options;
	gener_hf(hfw);
}

/*
void hfsubdiv2_callb(GtkWidget *wdg, gpointer data) {
	// Simple subdivision, triangles
	hf_wrapper_struct *hfw;
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	my_msg("HFSUBDIV2_CALLB",WARNING);
}
*/
void hfsurfadd_callb(GtkWidget *wdg, gpointer data) {
	// Progressive addition of surfaces

	hf_wrapper_struct *hfw;
	hf_type value;
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	if (!hfw)
		return;
	//  Options dialog - we create it at the 1st pass
	if (!hfw->hf_options->surfadd_dialog) {
	//	We create the struct which will receive the data
		hfw->hf_options->surfadd_options = surfadd_opt_new();
		hfw->hf_options->surfadd_options->dist_matrix = hfw->hf_options->dist_matrix;
	//	Dialog creation (with some defaults from surfadd_opt_new())
		hfw->hf_options->surfadd_dialog = surfadd_dialog_new(data);
		hfw->hf_options->surfadd_options->current_filter =
			&((hf_wrapper_struct *)hfw->hf_options->primit_surf)->hf_options->fd_struct->current_filter;
		hfw->hf_options->surfadd_options->revert_filter =
			&((hf_wrapper_struct *)hfw->hf_options->primit_surf)->hf_options->fd_struct->revert_filter;
		hfw->hf_options->surfadd_options->filter_merge =
			&((hf_wrapper_struct *)hfw->hf_options->primit_surf)->hf_options->fd_struct->merge_oper;
		hfw->hf_options->surfadd_options->filter_level =
			&((hf_wrapper_struct *)hfw->hf_options->primit_surf)->hf_options->fd_struct->filter_level;
		gtk_container_add(GTK_CONTAINER(hfw->creation_container),
			hfw->hf_options->surfadd_dialog);
	}
	if (hfw->hf_options->current_dialog)
		gtk_widget_hide(hfw->hf_options->current_dialog);
	gtk_widget_show(hfw->hf_options->surfadd_dialog);
	hfw->hf_options->last_creation_action = _("SurfSum");
	hfw->hf_options->current_dialog = hfw->hf_options->surfadd_dialog;
	hfw->hf_options->current_calculation = (gpointer) surfadd;
	hfw->hf_options->current_callb = hfsurfadd_callb;
	hfw->hf_options->current_options = (gpointer) hfw->hf_options->surfadd_options;

	gener_hf(hfw);
}

void hfwaves_callb(GtkWidget *wdg, gpointer data) {
	hf_wrapper_struct *hfw;
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	my_msg("HFWAVES_CALLB",WARNING);
}

void hfwaves_circ_callb(GtkWidget *wdg, gpointer data) {
	hf_wrapper_struct *hfw;
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	my_msg("HFWAVES_CIRC_CALLB",WARNING);
}

//	HF tools

void transform_callb(GtkWidget *wdg, gpointer data) {
	hf_wrapper_struct *hfw;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	if (!hfw) return;
// printf("ENTERING TRANSFORM\n");
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg))) {
		// We are leaving the "transform" context, commit the changes
// printf("LEAVING TRANSFORM\n");
			hfw->hf_options->need_tmp_buf = FALSE;
			hf_commit_or_reset (hfw->hf_options);
			return;
		}
	commit_pending_record(hfw);
	hfw->hf_options->current_cursor = hfw->hf_options->default_cursor;
	gdk_window_set_cursor(hfw->area->window, hfw->hf_options->current_cursor);
	if (hfw->hf_options->current_tools)
		gtk_widget_hide(hfw->hf_options->current_tools);
	if (hfw->hf_options->img) {
		hfw->hf_options->current_tools = hfw->hf_options->img->img_dialog;
		gtk_widget_show(hfw->hf_options->current_tools);
	}
	hf_backup(hfw->hf_struct);
	hfw->hf_options->need_tmp_buf = TRUE;
	hfw->hf_options->context = TRANSFORM_CONTEXT;
}

void draw_tools_callb(GtkWidget *wdg, gpointer data) {
	hf_wrapper_struct *hfw;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	if (!hfw) return;
// printf("ENTERING DRAW_TOOLS\n");
	if (wdg)
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg))) {
		// We are leaving the "draw" context, commit the changes
// printf("LEAVING DRAW_TOOLS\n");
			hfw->hf_options->need_tmp_buf = FALSE;
			hf_commit_or_reset (hfw->hf_options);
			return;
		}
	commit_pending_record(hfw);
	hfw->hf_options->current_cursor = hfw->hf_options->pencil_cursor;
	gdk_window_set_cursor(hfw->area->window, hfw->hf_options->current_cursor);
	if (hfw->hf_options->current_tools)
		gtk_widget_hide(hfw->hf_options->current_tools);
	if (hfw->hf_options->draw) {
		if (hfw->hf_options->draw->draw_tools_dialog) {
			hfw->hf_options->current_tools = hfw->hf_options->draw->draw_tools_dialog;
			gtk_widget_show(hfw->hf_options->current_tools);
		}
		else
			return;
	}
	hf_backup(hfw->hf_struct);
	hfw->hf_options->need_tmp_buf = TRUE;
	hfw->hf_options->context = DRAW_TOOLS_CONTEXT;
}

void render_pov(hf_wrapper_struct *hfw) {
	static char *buf[] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};
	gchar *fname;
	gint i;
	fname = get_file_name (*hfw->filename, FILESEP);
	// Remove spaces
	i = strlen(fname)-1;
	while ((fname[i]==' ') && (i>0)) i--;
	fname[i+1]='\0';
//	printf("FNAME: %d - %s; FILENAME: %d - %s; %d\n",strlen(fname), fname, strlen(*hfw->filename), *hfw->filename, fname[10]);
	// 2004-12-23 Arguments separated for compatibility with MDK10.1
	buf[0] = RENDERER;
	buf[1] = (gchar *) malloc(strlen(SCENE)+5);
	buf[2] = (gchar *) malloc(15);
	buf[3] = (gchar *) malloc(15);
	buf[4] = (gchar *) malloc(strlen(fname)+strlen(OUTPUT_PREFIX)+5);
	buf[5] =  OTHER_RENDER_OPTIONS;
	sprintf(buf[1],"+I%s",SCENE);
	sprintf(buf[2],"+W%05d",RENDER_WIDTH);
	sprintf(buf[3],"+H%05d",RENDER_HEIGHT);
	sprintf(buf[4],"+O%s%s",OUTPUT_PREFIX,fname);
// printf("CURRENT_DIR: %s\n",get_current_dir_name());
// printf("POV: str: %s; W: %s; H: %s; Output: %s; RENDER_OPTIONS: %s\n",buf[1],buf[2],buf[3], buf[4], buf[5]);
	hfw->hf_options->povray_pid = execution(buf[0],buf);
//	printf("PID: %d\n",hfw->hf_options->povray_pid);
	if (buf[1]) {
		free(buf[1]);
		buf[1]=NULL;
	}
	if (buf[2]) {
		free(buf[2]);
		buf[2]=NULL;
	}
	if (buf[3]) {
		free(buf[3]);
		buf[3]=NULL;
	}
	if (buf[4]) {
		free(buf[4]);
		buf[4]=NULL;
		if (fname)
			free(fname);
	}
}

void run_callb(GtkWidget *wdg, gpointer data) {
//	Render a povray image, display only
//	First we save the image under HF_OUTPUT_FOR_RENDERING
//	(usually test.png) in the current directory
	gchar *dir;
	gfloat ar;
	hf_wrapper_struct *hfw;
	if (wdg)
		if (GTK_IS_TOGGLE_BUTTON(wdg))
			if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
				return;
	if (!data) return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	if (!hfw) return;
//	commit_pending_record(hfw);
//	hfw->hf_options->current_cursor = gdk_cursor_new(GDK_LEFT_PTR);
// printf("OUTPUT: %s\n",HF_OUTPUT_FOR_RENDERING);
	dir = (gchar *) get_current_dir_name();
	chdir(POV_DIR);
	hf_wrapper_save(hfw, HF_OUTPUT_FOR_RENDERING);
	
	// Patch 2005-05 - Deduce aspect ratio from image size,
	// backup the original (which is the preview W/H ratio), maybe for the future...
	ar = hfw->gl_preview->cameras[hfw->gl_preview->current_camera_id]->aspect_ratio;
	hfw->gl_preview->cameras[hfw->gl_preview->current_camera_id]->aspect_ratio = ((gfloat) RENDER_WIDTH) / (gfloat) RENDER_HEIGHT;
	
	gl_save_camera_inc (hfw->gl_preview->cameras[hfw->gl_preview->current_camera_id]);
	hfw->gl_preview->cameras[hfw->gl_preview->current_camera_id]->aspect_ratio = ar;
	render_pov(hfw);
	chdir(dir);
	free(dir);
}

void run_with_choice_callb(GtkWidget *wdg, gpointer data) {
//	Render a povray image after choosing a scene file
	gchar *path_n_file;
	if (wdg)
		if (GTK_IS_TOGGLE_BUTTON(wdg))
			if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
				return;
	path_n_file =
		GetFilename(	scene_ext_list,
				NB_SCENE_EXT,
				_("Open"),
				POV_DIR,
				EXISTING_FILE,
				scene_ext_list[0].lbl);
	if (path_n_file)
		SCENE = path_n_file;
	else
		return;
	POV_DIR = get_dir_name(path_n_file,FILESEP);
	// It could be dangerous to free SCENE before reinitializing it
	// It could contain a default #define value (check this... small memory leakage here).
	// However path_n_file is always "malloced" in GetFilename
	run_callb(wdg, data);
}

void stop_callb(GtkWidget *wdg, gpointer data) {
	hf_wrapper_struct *hfw;
	if (!wdg)
		return;
	if (GTK_IS_TOGGLE_BUTTON(wdg))
		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wdg)))
			return;
	if (!data)
		return;
	hfw = (hf_wrapper_struct *) * (hf_wrapper_struct **) data;
	if (!hfw)
		return;
	if (hfw->hf_options->povray_pid<=0)
		return;
	kill((pid_t) hfw->hf_options->povray_pid, SIGKILL);
	hfw->hf_options->povray_pid = 0;
}

/**********************************************************************************************/

// Utilities requiring hf_wrapper_struct *

void set_watch_cursor (hf_wrapper_struct *hfw) {
	// For long processes...
	// We set it on all widgets / windows
	// The HF area has its own cursor (current->cursor)
	gdk_window_set_cursor(hfw->area->window, hfw->hf_options->watch_cursor);
	gdk_window_set_cursor(hfw->window->window, hfw->hf_options->watch_cursor);
	gdk_window_set_cursor(hfw->hf_options->tools_dialog->window, hfw->hf_options->watch_cursor);
	gdk_flush();
}

void unset_watch_cursor (hf_wrapper_struct *hfw) {
	gdk_window_set_cursor(hfw->area->window, hfw->hf_options->current_cursor);
	gdk_window_set_cursor(hfw->window->window, hfw->hf_options->default_cursor);
	gdk_window_set_cursor(hfw->hf_options->tools_dialog->window, hfw->hf_options->default_cursor);
}
