/* hf_dialog.h - headers for height field dialog class
 *
 * Copyright (C) 2001 Patrice St-Gelais
 *         patrstg@users.sourceforge.net
 *         www.oricom.ca/patrice.st-gelais
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

//	This file includes the prototypes for the dialog creation functions
//	for each HF subtypes
//	These functions are define in other files than hf_dialog.c

#ifndef _HF_DIALOG
#define _HF_DIALOG 1

#include <gtk/gtk.h>
#include "../utils/menus_n_tools.h"
#include "globals.h"
#include "hf.h"
#include "hf_filters.h"
#include "hf_filters_dialog.h"
#include "subdiv1.h"
#include "subdiv2.h"
#include "uniform.h"
#include "surfadd.h"
#include "img_process_dialog.h"
#include "draw_tools_dialog.h"
#include "gl_preview.h"
#include "view_area.h"
#include "dialog_utilities.h"
#include <unistd.h>

// #include "waves.h"

//	Scale labels
#define SCALE_12_5 "12,5"
#define SCALE_25 "25"
#define SCALE_50 "50"
#define SCALE_100 "100"
#define SCALE_200 "200"
#define SCALE_400 "400"

#ifndef SCALE_PAD
#define SCALE_PAD 25
#endif

//	HF types:  pen (for painting) or main (background)
#ifndef HFPEN
	#define HFPEN 1
#endif
#ifndef HFMAIN
	#define HFMAIN 2
#endif
//	This one is for progressive surface addition
#ifndef HFPRIMITIVE
	#define HFPRIMITIVE 3
#endif

//	For adding noise by merging a simple subdivision with another HF
#ifndef HFNOISE
	#define HFNOISE 3
#endif

//	Context (dialog ID)

#ifndef DRAW_TOOLS_CONTEXT
	#define DRAW_TOOLS_CONTEXT 1
#endif
#ifndef TRANSFORM_CONTEXT
	#define TRANSFORM_CONTEXT 2
#endif

typedef struct {
	GtkWidget *hf_types_toolbar;
	omenu_struct_type *size_menu;
} hf_creation_wdg_struct;

typedef struct {
//	HF CREATION OPTIONS
//	A few options inherited from the tools dialog
//	For building new widgets, for instance toolbars
	GtkWidget *tools_window;
	GtkWidget *tools_dialog;		// Where we put the tools dialog
					// Inherited from the doc_type struct
					// A vbox to which we can add / subtract widgets
	hf_creation_wdg_struct *creation_widgets;
	GtkWidget *current_tools;		// Current tools dialog, to show / hide
	GtkTooltips *tooltips;
	GtkWidget *current_dialog;	// Current creation subdialog
	gint context;			// Current context (dialog ID) - most important for drawing
	gchar *last_creation_action;	// String describing the current creation dialog
						// It gives the last action performed on a new document,
						// recorded in the undo / redo stack
//	Other options
	GtkWidget *uniform_dialog;
	hf_type *uniform_value;
	gboolean flag_uniform_done;  // True if the HF has been committed with the filter
				// Avoid recomputing the HF when drawing (faster!)
	GtkWidget *subdiv1_dialog;
	subdiv1_opt *subdiv1_options;
	GtkWidget *subdiv2_dialog;
	GtkObject *freq_subdiv2_adj[12];	// Adjustments for frequency control, from [0]=2 to [11]=4096
	gboolean subdiv2_direct_upd;
	subdiv2_opt *subdiv2_options;
//	gpointer subdiv2_options;
	GtkWidget *surfadd_dialog;
	surfadd_opt *surfadd_options;
	GtkWidget *waves_dialog;
//	waves_opt *waves_options;
	gpointer waves_options;
	void (*current_callb) (GtkWidget *wdg, gpointer data);  // For executing default options
	gpointer current_options;	// Can be *uniform_value, subdiv1-2_options, sufadd_options....
	gpointer (*current_calculation) (gpointer hf, gpointer options); // Calculates HF
					// Varies with kind of HF
//	Shape filters defined once in the application
	dist_matrix_struct *dist_matrix;
	filter_struct *wide_filter;
	filter_struct *medium_filter1;
	filter_struct *medium_filter2;
	filter_struct *sharp_filter;
	filter_struct *gauss_filter;
	filter_struct *image_filter;
//	Filter dialog struct for creation (there is another one in ->img)
	filter_dialog_struct *creation_fd_struct;
//	Current filter struct we are processing (= ->creation_fd_struct or ->img->filter_dialog)
	filter_dialog_struct *fd_struct;

//	List of opened HF (required by the filter dialog, and maybe others in the future)
	GList *doc_list;

	draw_tools_struct *draw;

	GdkCursor *current_cursor; // Current cursor for the document window (pen / arrow)
	GdkCursor *default_cursor;
	GdkCursor *pencil_cursor;
	GdkCursor *watch_cursor;
	GdkCursor *fleur_cursor;
	gint hf_size;	// Result of the size widget, in the creation dialog
	gpointer primit_surf;	// HF wrapper for primitive surface (SURFADD)
	GtkObject *freq_control_adj[12];	// Adjustments for frequency control, from [0]=2 to [11]=4096
	img_dialog_struct *img;		// Dialogs and structures for image transformations
	gboolean need_tmp_buf;		// TRUE if we need to initialize hf_struct->tmp_buf after an undo
	gl_defaults_struct *gl_defaults;	// For OpenGL preview display
	gdouble *gauss_list[GAUSS_LIST_LENGTH];	// Cache of normalized gaussian vectors, for drawing
	gpointer (**commit_fn) (gpointer);	// Address of the pointer of the committing function
					// Executed when changing context
					// This is a pointer to a function pointer, known to the global application

	gpointer (**reset_fn) (gpointer);
	gpointer commit_data; // Pointer to the data pointer required by commit_fn
					// This is also a bucket known to the global application
	gint povray_pid; // Last Povray launch
} hf_options_struct;	// Optional parts of the creation dialog

typedef struct {
	gchar **filename;			// Inherited from the embedding document structure
					// Needed for manipulating list of opened HF
	GtkWidget *window;		// HF main window
	GtkWidget *area;
	guint exp_handler;		// Connect handler to "area" expose event
	gint area_size;			// Size of area (supposed to be square, avoid calculation)
	gl_preview_struct *gl_preview;
	control_line_struct *control_line;
	GtkWidget *creation_container;	// Where we put the hf options during creation
					// Inherited from the doc_type_struct
	GtkWidget *tiling_wdg;		// A checkbox connected to hf_struct->if_tiles
	gint type;			// HFPEN, HFMAIN, HFPRIMITIVE
	hf_struct_type *hf_struct;		// Contains the hf data
	hf_options_struct *hf_options;  // Redundancy of optional parts of the creation dialog
				// The original is in doc_type_struct as a gpointer
	gboolean mouse_dragging;	// Used for mouse tracking
	GtkWidget *mouse_position;	// A label containing (mousex,mousey)
	gboolean creation_mode;		// TRUE if creation not commited
	gboolean if_calculated;		// For managing recomputing from options
	gboolean apply_filter;		// TRUE:  we must apply the current filter
	gboolean *if_modified;		// "Bucket" inherited from the document wrapper
					// in the calling application structure.
					// Set to TRUE after any update
	gboolean *to_register_in_history;	// Flag used by the calling app for managing
					// the undo-redo stack;
					// must be set to TRUE after any significant update
	void (*init_history_flag) (GtkWidget *, gboolean *);  // Function which inits the preceding flag
	gchar *last_action;		// Last modif, info for undo / redo stack
	gint display_scale;		// Output of a widget controlling display scale
					// Transmitted (inherited to) hf_struct
	GSList *scale_group;		// List of scale buttons (widgets)
	GtkWidget *draw_area_frame;		// Frame containing the HF pixmap
	gpointer all_docs;		// Structure inherited from the calling application,
					// containing all the open documents
	gchar *doc_type;		// Name of the current document type,
					// inherited from the calling application
	GList *(*get_doc_list_from_type) (gpointer all_docs, gchar* doc_type);
					// Extracts all the documents of the current type,
					// for building menus, etc.   Inherited.
	gboolean size_change_request;	// For managing expose and configure events after
					// a size change request
} hf_wrapper_struct;

GtkWidget* hf_creation_dialog_new (GtkWidget *window, GtkTooltips *tooltips,
			gpointer callb_data, gpointer widget_list_ptr);

GtkWidget* hf_tools_dialog_new (GtkWidget *window, GtkTooltips *tooltips,
	gpointer hf_options, gpointer callb_data);

GtkWidget *hf_types_toolbar_new (GtkWidget *window, GtkTooltips *tooltips,
		GtkWidget **tb_widget, gpointer callb_data);

gpointer hf_options_dialog_new (GtkWidget*, GtkWidget *, GtkTooltips*,
	gpointer widget_list, gpointer callback_data,
	gpointer commit_fn_ptr,
	gpointer reset_fn_ptr,
	gpointer commit_data_ptr);

void hf_options_dialog_free(hf_options_struct *);

void set_commit_reset_fns (hf_options_struct *,
	gpointer commit_fn,
	gpointer reset_fn,
	gpointer data );

gpointer hf_wrapper_new (gchar **filename, GtkWidget *window, GtkWidget *creation_dialog,
		gpointer callback_data,
		gboolean *if_modified, gboolean *to_register_in_history,
		void (*init_history_flag) (GtkWidget *, gboolean*),
		gchar *path_n_file, gboolean *fname_tochoose,
		gchar *doc_type, gpointer all_docs,
		GList *(*get_doc_list_from_type) (gpointer all_docs, gchar *doc_type));
void hf_wrapper_free(gpointer, gboolean);
gpointer hf_wrapper_copy(gpointer hfw_from, gpointer hfw_to);
void hf_wrapper_display(gpointer hfw_struct);
void draw_hf (hf_wrapper_struct *) ;
void draw_hf_partial (hf_wrapper_struct *hfw,
		gint from_x, gint to_x, gint from_y, gint to_y,
		gboolean update_gl_area) ;
void gener_hf (hf_wrapper_struct *hfw);
gchar* hf_get_last_action (gpointer);
gint hf_default_size();
void hf_wrapper_save (gpointer data, gchar *path_n_file);
void hf_stats (gpointer hf_wrapper);
gint get_display_scale(gint hf_size);
void set_widget_display_scale (hf_wrapper_struct *);

void hfuniform_callb(GtkWidget *wdg, gpointer data) ;
void hfsubdiv1_callb(GtkWidget *wdg, gpointer data);
void hfsubdiv2_callb(GtkWidget *wdg, gpointer data);
void hfsurfadd_callb(GtkWidget *wdg, gpointer data);
// void hfwaves_callb(GtkWidget *wdg, gpointer data);
// void hfwaves_circ_callb(GtkWidget *wdg, gpointer data);
void nofilter_callb(GtkWidget *wdg, gpointer data) ;
void widefilter_callb(GtkWidget *wdg, gpointer data);
void mediumfilter1_callb(GtkWidget *wdg, gpointer data);
void mediumfilter2_callb(GtkWidget *wdg, gpointer data);
void sharpfilter_callb(GtkWidget *wdg, gpointer data);

void draw_tools_callb(GtkWidget *,gpointer );
void transform_callb(GtkWidget *, gpointer );

void run_callb(GtkWidget *, gpointer );
void run_with_choice_callb(GtkWidget *, gpointer );
void stop_callb(GtkWidget *, gpointer );

void record_action(hf_wrapper_struct *, gchar *action);
void begin_pending_record(hf_wrapper_struct *, 
	gchar *action,
	gpointer commit_fn,
	gpointer reset_fn);
void commit_pending_record(hf_wrapper_struct *);
void hf_commit_or_reset (hf_options_struct *hfo);
void accept_fn (hf_wrapper_struct *hfw);
void reset_fn (hf_wrapper_struct *hfw);
void accept_fault_fn (hf_wrapper_struct *hfw);

void hf_commit_creation(gpointer);

void set_watch_cursor (hf_wrapper_struct *hfw);
void unset_watch_cursor (hf_wrapper_struct *hfw);

GtkWidget * hf_get_icon (gpointer hfw, gint size_x, gint size_y);

/**************************************************************************************************/

//	Prototypes for HF subtypes creation dialog functions
GtkWidget * uniform_dialog_new(gpointer data, gint default_value);

GtkWidget *subdiv1_dialog_new(gpointer data);
GtkWidget *subdiv2_dialog_new(gpointer data);
void set_frq_vector(hf_wrapper_struct *hfw,gint *frq, gint nbrfrq);	// For setting frequencies in subdiv2
GtkWidget *surfadd_dialog_new(gpointer data);
// GtkWidget *waves_dialog_new(gpointer data);

//	Public methods from gl_preview.c

void gl_area_update(hf_wrapper_struct *hfw) ;
gl_defaults_struct *gl_defaults_init( );
gl_preview_struct *create_preview_dialog (gl_defaults_struct *gl_def, gpointer callb_data);

#endif // _HF_DIALOG
