/* img_process.h - headers for image processing filters
 *
 * Copyright (C) 2001 Patrice St-Gelais
 *         patrstg@users.sourceforge.net
 *         www.oricom.ca/patrice.st-gelais
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _IMGPROCESS
#define _IMGPROCESS 1

#include "globals.h"
#include "hf.h"
#include "merge_dialog.h"

typedef struct {
	gint x;
	gint y;
	hf_type val;
} x_y_val;

void hf_brightness_contrast(	hf_struct_type *hf, 
				gint contrast_level, 
				gint brightness_level,
				gboolean keep_luminosity,
				gboolean bound_overflow);
void hf_auto_contrast (hf_struct_type *hf);
void hf_terrace (hf_struct_type *hf, gint levels, gint percent, gint seed, gint percent_random,
	gint merge, gint source_offset, gint result_offset, gint smooth_radius, gboolean if_wrap,
	dist_matrix_struct *dist_matrix, gint artifact_radius, 
	gboolean apply_postprocess, gboolean merge_change,
	gint merge_x_translate, gint merge_y_translate, gdouble **gauss_list) ;
void hf_threshold (hf_struct_type *hf, hf_type min, hf_type max, gint percent);
void hf_smooth (hf_struct_type *hf, gint radius, dist_matrix_struct *, gboolean wrap, gdouble **gauss_list);
void hf_city(hf_struct_type *hf, gint skycraper_width, gint skyscraper_var, gint streets_width, merge_struct *mrg);
void hf_simple_blur(hf_struct_type *hf, gint radius);
hf_type * hf_subtract(hf_type *hf1, hf_type *hf2, gint length);
void hf_stretch_v(hf_struct_type *hf);
void hf_stretch_h(hf_struct_type *hf);
void hf_convolve (hf_struct_type *hf, gint radius, gdouble *matrix, gboolean if_tiles);
void hf_erode (hf_struct_type *hf, gint steps);
void hf_remove_artifacts (hf_struct_type *hf, gint radius, dist_matrix_struct *dm);
void hf_hexagon (hf_struct_type *hf, gint radius, gint border, gint mix,
	gint merge_op, gint source_offset, gint result_offset, 
	gint smooth_radius, 
	dist_matrix_struct *dist_matrix,
	gboolean apply_postprocess,
	gboolean merge_change,
	gint merge_x_translate,
	gint merge_y_translate,
	gdouble **gauss_list);

void hf_math_fn (hf_struct_type *hf, gint math_op, gdouble parameter);

void hf_map_convolve (hf_struct_type *map, hf_struct_type *background, gint x, gint y,
	gboolean wrap, gint level, gdouble **gauss_list);

#endif // _IMGPROCESS



